from unittest import TestCase
import os
import numpy as np
from numpy.testing import assert_array_equal
import pyilastik
import pyilastik.ilastik_storage_version_01 as ils

data_path = path = os.path.join(os.path.dirname(__file__), 'data')


val_multi_z = np.array([[[0., 0., 0., 0., 0., 0., 0., 0., 0., 0.],
                         [0., 0., 1., 1., 0., 0., 0., 0., 0., 0.],
                         [0., 0., 2., 2., 0., 0., 0., 0., 0., 0.],
                         [0., 0., 0., 0., 0., 0., 0., 0., 0., 0.],
                         [0., 0., 0., 0., 0., 0., 0., 0., 0., 0.],
                         [0., 0., 0., 0., 0., 0., 0., 0., 0., 0.],
                         [0., 0., 0., 0., 0., 0., 0., 0., 1., 1.],
                         [0., 0., 0., 0., 0., 0., 0., 0., 2., 2.]],
                        [[0., 0., 0., 0., 0., 0., 0., 0., 0., 0.],
                         [0., 0., 0., 0., 0., 0., 0., 0., 0., 0.],
                         [0., 0., 0., 0., 0., 0., 0., 0., 0., 0.],
                         [0., 0., 0., 0., 1., 0., 0., 0., 0., 0.],
                         [0., 0., 0., 0., 0., 0., 0., 0., 0., 0.],
                         [0., 0., 0., 0., 0., 0., 0., 0., 0., 0.],
                         [0., 0., 0., 0., 0., 0., 0., 0., 0., 0.],
                         [0., 0., 0., 0., 0., 0., 0., 0., 0., 0.]]])

val_single_z = val_multi_z[0, :, :]

# swapped label values
val_multi_z_swapped = np.zeros(val_multi_z.shape)
val_multi_z_swapped[val_multi_z == 2] = 1
val_multi_z_swapped[val_multi_z == 1] = 2

val_single_z_swapped = val_multi_z_swapped[0, :, :]


class TestLabelImportDimensions(TestCase):

    def test_single_channel_single_z(self):

        p = os.path.join(path, 'dimensionstest/x15_y10_z1_c1_classes2.ilp')
        ilp = pyilastik.read_project(p, skip_image=True)
        (_, (_, labels, _)) = ilp['x15_y10_z1_c1.tif']

        self.assertEqual(labels.shape, (1, 8, 10, 1))
        assert_array_equal(labels[0, :, :, 0], val_single_z)

    def test_single_channel_multi_z(self):

        p = os.path.join(path, 'dimensionstest/x15_y10_z2_c1_classes2.ilp')
        ilp = pyilastik.read_project(p, skip_image=True)
        (_, (_, labels, _)) = ilp['x15_y10_z2_c1.tif']

        self.assertEqual(labels.shape, (2, 8, 10, 1))
        assert_array_equal(labels[:, :, :, 0], val_multi_z)

    def test_single_channel_multi_z_2(self):

        p = os.path.join(path, 'ilastik-1.2.ilp')
        # 2x 4y 8z
        ilp = pyilastik.read_project(p, skip_image=True)
        (_, (_, labels, _)) = ilp['ilastik-test-2-4-8.tif']

        self.assertEqual(labels.shape, (8, 4, 2, 1))

    def test_multi_channel_multi_z(self):

        p = os.path.join(path, 'dimensionstest/x15_y10_z2_c4_classes2.ilp')
        ilp = pyilastik.read_project(p, skip_image=True)
        (_, (_, labels, _)) = ilp['x15_y10_z2_c4.tif']

        self.assertEqual(labels.shape, (2, 8, 10, 1))
        assert_array_equal(labels[:, :, :, 0], val_multi_z)

        # print(np.transpose(labels, (3, 0, 2, 1)).astype(int))
        # assert False


    def test_caching_of_block_slices(self):

        p = os.path.join(path, 'dimensionstest/x15_y10_z2_c4_classes2.ilp')
        ilp = pyilastik.read_project(p, skip_image=True)

        s = np.array([[0, 2], [0, 10], [0, 15], [0, 1]])

        bs_before = ilp._get_block_slices(0).copy()
        tile = ilp.tile(0, s)
        bs_after = ilp._get_block_slices(0)
        print(bs_before)
        assert_array_equal(bs_before, bs_after)







    def test_multi_channel_single_z(self):

        p = os.path.join(path, 'dimensionstest/x15_y10_z1_c2_classes2.ilp')
        ilp = pyilastik.read_project(p, skip_image=True)
        (_, (_, labels, _)) = ilp['x15_y10_z1_c2.tif']

        self.assertEqual(labels.shape, (1, 8, 10, 1))
        assert_array_equal(labels[0, :, :, 0], val_single_z)

    def test_rgb_single_z(self):

        p = os.path.join(path, 'dimensionstest/x15_y10_z1_rgb_classes2.ilp')
        ilp = pyilastik.read_project(p, skip_image=True)
        (_, (_, labels, _)) = ilp['x15_y10_z1_rgb.tif']

        self.assertEqual(labels.shape, (1, 8, 10, 1))
        assert_array_equal(labels[0, :, :, 0], val_single_z_swapped)

    def test_rgb_mutli_z(self):

        p = os.path.join(path, 'dimensionstest/x15_y10_z2_rgb_classes2.ilp')
        ilp = pyilastik.read_project(p, skip_image=True)
        (_, (_, labels, _)) = ilp['x15_y10_z2_rgb.tif']

        self.assertEqual(labels.shape, (2, 8, 10, 1))
        assert_array_equal(labels[:, :, :, 0], val_multi_z)

    def test_shape_of_labelmatrix_large_dat(self):

        localpath = os.path.join(path, 'purkinjetest')
        p = os.path.join(localpath, 'ilastik-1.2.2post1mac.ilp')
        ilp = pyilastik.read_project(p, skip_image=True)

        # 758_cerebellum_5M60_subset_3.tif
        mat_shape = ilp.shape_of_labelmatrix(1)
        assert_array_equal(mat_shape, np.array((684, 861, 1)))

    def test_labels_large_dat(self):

        localpath = os.path.join(path, 'purkinjetest')
        p = os.path.join(localpath, 'ilastik-1.2.2post1mac.ilp')
        ilp = pyilastik.read_project(p, skip_image=True)
        (_, (_, labels, _)) = ilp['images/769_cerebellum_5M41_subset_1.tif']

        val = np.array([[2., 2., 2., 2.],
                        [0., 2., 2., 2.],
                        [0., 0., 2., 2.],
                        [0., 0., 4., 4.],
                        [4., 4., 4., 4.]])

        assert_array_equal(val, labels[0, 212:217, 309:313, 0])

    def test_read_project_for_ilastik_version_1_3_0(self):
        p = os.path.join(path, 'ilastik-1.3.ilp')
        ilp = pyilastik.read_project(p, skip_image=True)
        labels = ilp['ilastik-test-2-4-8.tif'][1][1]

        assert labels.shape == (8, 4, 2, 1)
        assert set(np.unique(labels)) == {0, 1, 2}

    def test_get_block_slices(self):

        localpath = os.path.join(path, 'purkinjetest')
        p = os.path.join(localpath, 'ilastik-1.2.2post1mac.ilp')

        #p = os.path.join(path, 'dimensionstest/x15_y10_z2_rgb_classes2.ilp')
        ilp = pyilastik.read_project(p, skip_image=True)

        b = ilp._get_block_slices(0)
        val = np.array([[[253, 267],
                         [318, 484],
                         [0, 1]],
                        [[186, 253],
                         [289, 486],
                         [0, 1]]])
        assert_array_equal(b, val)


    def test_blocks_in_tile(self):


        localpath = os.path.join(path, 'purkinjetest')
        p = os.path.join(localpath, 'ilastik-1.2.2post1mac.ilp')

        tile_slices = np.array([[255, 300],
                                [312, 500],
                                [0, 1]])

        ilp = pyilastik.read_project(p, skip_image=True)
        b = ilp._blocks_in_tile(0, tile_slices)

        self.assertTrue(b[0])
        self.assertFalse(b[1])
        self.assertEqual(len(b), 2)


        p = os.path.join(path, 'dimensionstest/x502_y251_z5_c1_classes2.ilp')
        ilp = pyilastik.read_project(p, skip_image=True)

        tile_slices = np.array([[0, 1],
                               [0, 50],
                               [0, 50],
                               [0, 1]])
        assert_array_equal(ilp._blocks_in_tile(0, tile_slices),
                           np.array([True, False, False]))






    def test_tile(self):

        p = os.path.join(path, 'dimensionstest/x502_y251_z5_c1_classes2.ilp')
        ilp = pyilastik.read_project(p, skip_image=True)

        tile_slice = np.array([[0, 1],
                               [0, 10],
                               [0, 7],
                               [0, 1]])

        t = ilp.tile(0, tile_slice)

        val_1 = np.array([[0., 0., 0., 0., 0., 0., 0.],
                          [0., 0., 0., 0., 0., 0., 0.],
                          [0., 0., 0., 0., 0., 0., 0.],
                          [0., 0., 0., 0., 0., 0., 0.],
                          [0., 0., 1., 1., 1., 0., 0.],
                          [0., 0., 1., 1., 1., 0., 0.],
                          [0., 0., 1., 1., 1., 0., 0.],
                          [0., 0., 1., 1., 1., 0., 0.],
                          [0., 0., 0., 0., 0., 0., 0.],
                          [0., 0., 0., 0., 0., 0., 0.]])

        assert_array_equal(t[0, :, :, 0], val_1)

        tile_slice = np.array([[0, 1],
                               [92, 101],
                               [157, 164],
                               [0, 1]])

        t = ilp.tile(0, tile_slice)

        val_2 = np.array([[0., 0., 0., 0., 0., 0., 0.],
                          [0., 0., 0., 0., 0., 0., 0.],
                          [0., 0., 1., 1., 1., 1., 0.],
                          [0., 0., 1., 1., 1., 1., 0.],
                          [0., 0., 2., 2., 0., 0., 0.],
                          [0., 0., 2., 2., 0., 0., 0.],
                          [0., 0., 2., 2., 0., 0., 0.],
                          [0., 0., 2., 2., 0., 0., 0.],
                          [0., 0., 0., 0., 0., 0., 0.]])

        assert_array_equal(t[0, :, :, 0], val_2)

        tile_slice = np.array([[0, 1],
                               [50, 55],
                               [50, 55],
                               [0, 1]])

        t = ilp.tile(0, tile_slice)

        val_3 = np.array([[0., 0., 0., 0., 0.],
                          [0., 0., 0., 0., 0.],
                          [0., 0., 0., 0., 0.],
                          [0., 0., 0., 0., 0.],
                          [0., 0., 0., 0., 0.]])

        assert_array_equal(t[0, :, :, 0], val_3)




    def test_tile_for_selected_blocks(self):

        p = os.path.join(path, 'dimensionstest/x502_y251_z5_c1_classes2.ilp')
        ilp = pyilastik.read_project(p, skip_image=True)

        pos, tile = ilp.tile_for_selected_blocks(0, [True, False, True])

        tile_1 = tile[0, :10, :10, 0]
        tile_2 = tile[0, -10:, -10:, 0]

        self.assertEqual(tile.shape, (2, 96, 161, 1))

        val_1 = np.array([[1., 1., 1., 0., 0., 0., 0., 0., 0., 0.],
                          [1., 1., 1., 0., 0., 0., 0., 0., 0., 0.],
                          [1., 1., 1., 0., 0., 0., 0., 0., 0., 0.],
                          [1., 1., 1., 0., 0., 0., 0., 0., 0., 0.],
                          [0., 0., 0., 0., 0., 0., 0., 0., 0., 0.],
                          [0., 0., 0., 0., 0., 0., 0., 0., 0., 0.],
                          [0., 0., 0., 0., 0., 0., 0., 0., 0., 0.],
                          [0., 0., 0., 0., 0., 0., 0., 0., 0., 0.],
                          [0., 0., 0., 0., 0., 0., 0., 0., 0., 0.],
                          [0., 0., 0., 0., 0., 0., 0., 0., 0., 0.]])

        val_2 = np.array([[0., 0., 0., 0., 0., 0., 0., 0., 0., 0.],
                          [0., 0., 0., 0., 0., 0., 0., 0., 0., 0.],
                          [0., 0., 0., 0., 0., 0., 0., 0., 0., 0.],
                          [0., 0., 0., 0., 0., 0., 0., 0., 0., 0.],
                          [0., 0., 0., 0., 0., 0., 1., 1., 1., 1.],
                          [0., 0., 0., 0., 0., 0., 1., 1., 1., 1.],
                          [0., 0., 0., 0., 0., 0., 2., 2., 0., 0.],
                          [0., 0., 0., 0., 0., 0., 2., 2., 0., 0.],
                          [0., 0., 0., 0., 0., 0., 2., 2., 0., 0.],
                          [0., 0., 0., 0., 0., 0., 2., 2., 0., 0.]])

        assert_array_equal(tile_1, val_1)
        assert_array_equal(tile_2, val_2)






    def test_is_overlap(self):

        tpl = np.array([[2, 13],
                        [4, 15]])

        b1 = np.array([[3, 6],
                       [10, 14]])
        b2 = np.array([[9, 16],
                       [3, 7]])
        b3 = np.array([[13, 17],
                       [10, 13]])
        b4 = np.array([[0, 1],
                       [12, 13]])
        b5 = np.array([[1, 4],
                       [13, 17]])
        b6 = np.array([[3, 6],
                       [2, 5]])
        b7 = np.array([[12, 15],
                       [14, 17]])
        b8 = np.array([[0, 3],
                       [2, 5]])
        b9 = np.array([[0, 3],
                       [4, 7]])


        self.assertTrue(ils.is_overlap(tpl, b1))
        self.assertTrue(ils.is_overlap(tpl, b2))
        self.assertFalse(ils.is_overlap(tpl, b3))
        self.assertFalse(ils.is_overlap(tpl, b4))
        self.assertTrue(ils.is_overlap(tpl, b5))
        self.assertTrue(ils.is_overlap(tpl, b6))
        self.assertTrue(ils.is_overlap(tpl, b7))
        self.assertTrue(ils.is_overlap(tpl, b8))
        self.assertTrue(ils.is_overlap(tpl, b9))

        tpl = np.array([[2, 7],
                        [4, 10]])
        b1 = np.array([[2, 3],
                       [4, 10]])
        self.assertTrue(ils.is_overlap(tpl, b1))
        # assert False


    def test_tile_loc_from_slices(self):

        s1 = np.array([[3, 6],
                       [10, 14],
                       [0, 1]])

        s2 = np.array([[9, 16],
                       [3, 7],
                       [0, 1]])

        s3 = np.array([[12, 15],
                       [14, 17],
                       [0, 1]])

        pos, shape = ils.tile_loc_from_slices([s1, s2, s3])
        assert_array_equal(pos, np.array([3, 3, 0]))
        assert_array_equal(shape, np.array([13, 14, 1]))

    def test_get_slices_for(self):

        q_pos = np.array((5, 4))
        q_shape = np.array((3, 3))
        p_pos = np.array((4, 2))
        p_shape = np.array((5, 6))

        q_slice = ils._get_slices_for(p_pos, q_pos, p_shape, q_shape)
        p_slice = ils._get_slices_for(q_pos, p_pos, q_shape, p_shape)
        self.assertEqual(q_slice, [slice(0, 3), slice(0, 3)])
        self.assertEqual(p_slice, [slice(1, 4), slice(2, 5)])

        q_pos = np.array((5, 4))
        q_shape = np.array((7, 3))
        p_pos = np.array((4, 2))
        p_shape = np.array((5, 6))

        q_slice = ils._get_slices_for(p_pos, q_pos, p_shape, q_shape)
        p_slice = ils._get_slices_for(q_pos, p_pos, q_shape, p_shape)
        self.assertEqual(q_slice, [slice(0, 4), slice(0, 3)])
        self.assertEqual(p_slice, [slice(1, 5), slice(2, 5)])

        q_pos = np.array((1, 3))
        q_shape = np.array((7, 4))
        p_pos = np.array((4, 2))
        p_shape = np.array((5, 6))

        q_slice = ils._get_slices_for(p_pos, q_pos, p_shape, q_shape)
        p_slice = ils._get_slices_for(q_pos, p_pos, q_shape, p_shape)
        self.assertEqual(q_slice, [slice(3, 7), slice(0, 4)])
        self.assertEqual(p_slice, [slice(0, 4), slice(1, 5)])



import os
import numpy as np
from numpy.testing import assert_array_equal
import pyilastik

path = os.path.join(os.path.dirname(__file__), 'data')


def test_ilastik_version05():
    print(path)
    p = os.path.join(path, 'ilastik-0.5.ilp')
    for fname, (img, labels, prediction) in \
            pyilastik.read_project(p, image_path=path):
        np.testing.assert_equal(fname, 'ilastik-test-2-4-8.tif')
        np.testing.assert_array_equal(img.shape, (1, 1, 4, 2, 1))
        np.testing.assert_array_equal(img.shape, labels.shape)
        assert prediction is None


def test_ilastik_storage_version01():
    p = os.path.join(path, 'ilastik-1.2.ilp')
    for fname, (img, labels, prediction) in \
            pyilastik.read_project(p, skip_image=True):
        np.testing.assert_equal(fname, 'ilastik-test-2-4-8.tif')
        # np.testing.assert_array_equal(img.shape, (8, 4, 2, 1))
        # np.testing.assert_array_equal(img.shape, labels.shape)
        assert prediction is None


def test_2channels_1z_2classes():
    localpath = os.path.join(path, '2channels_1z_2classes')
    p = os.path.join(localpath, 'ilastik-1.2.2post1_mac.ilp')
    img_path = os.path.join(localpath, 'images')
    ilp = pyilastik.read_project(p, image_path=img_path, skip_image=True)
    (_, (_, labels, _)) = ilp['57x54_2channels_1z_2classes.tif']
    print(labels.shape)
    assert set(np.unique(labels[0, :, :4, 0])) == {1.}
    assert set(np.unique(labels[0, :, 4:, 0])) == {2.}


def test_1channel_1z_2classes():
    localpath = os.path.join(path, '1channel_1z_2classes')
    p = os.path.join(localpath, 'ilastik_1.2.2post1_mac.ilp')
    img_path = os.path.join(localpath, 'images')
    ilp = pyilastik.read_project(p, image_path=img_path, skip_image=True)
    (_, (_, labels, _)) = ilp['57x54_1channel_1z_2classes.tif']

    assert set(np.unique(labels[0, :18, :, 0][:])) == {2.}
    assert set(np.unique(labels[0, 18:, :21, 0][:])) == {1.}
    assert set(np.unique(labels[0, 18:, 21:, 0][:])) == {0.}


def test_shape_of_labelmatrix():
    localpath = os.path.join(path, '1channel_1z_2classes')
    p = os.path.join(localpath, 'ilastik_1.2.2post1_mac.ilp')
    ilp = pyilastik.read_project(p, skip_image=True)
    mat_shape = ilp.shape_of_labelmatrix(0)
    print(mat_shape)
    assert_array_equal(mat_shape, np.array((54, 57, 1)))

    localpath = os.path.join(path, '2channels_1z_2classes')
    p = os.path.join(localpath, 'ilastik-1.2.2post1_mac.ilp')
    ilp = pyilastik.read_project(p, skip_image=True)
    mat_shape = ilp.shape_of_labelmatrix(0)
    assert_array_equal(mat_shape, np.array((54, 57, 1)))

    p = os.path.join(path, 'ilastik-1.2.ilp')
    ilp = pyilastik.read_project(p, skip_image=True)
    mat_shape = ilp.shape_of_labelmatrix(0)
    assert_array_equal(mat_shape, np.array((8, 4, 2, 1)))


def test_ilastik0122_mac():

    localpath = os.path.join(path, 'purkinjetest')
    p = os.path.join(localpath, 'ilastik-1.2.2post1mac.ilp')
    img_path = os.path.join(localpath, 'images')
    ilp = pyilastik.read_project(p, image_path=img_path, skip_image=True)

    (_, (_, labels, _)) = ilp['images/769_cerebellum_5M41_subset_1.tif']
    assert set(np.unique(labels[:])) == {0, 1, 2, 3, 4}  # all 4 labels

    (_, (_, labels, _)) = ilp['images/758_cerebellum_5M60_subset_3.tif']
    assert set(np.unique(labels[:])) == {0, 1, 2, 3, 4}  # all 4 labels

    (_, (_, labels, _)) = ilp['images/667_cerebellum_5M53_subset_3.tif']
    print(np.unique(labels[:]))
    assert set(np.unique(labels[:])) == set([])  # all no labels

    (_, (_, labels, _)) = ilp['images/628_cerebellum_5M46_subset_1.tif']
    assert set(np.unique(labels[:])) == {0, 2, 3, 4}  # without label 1



import re
import pyilastik.utils as utils
import numpy as np

# image shape: (?,?,H,W,C), e.g. (1, 1, 2098, 2611, 3)
# labels shape: (?,?,H,W,1), e.g. (1, 1, 2098, 2611, 1), 0 == unlabeled
# prediction shape: (?,?,H,W,L), e.g. (1, 1, 2098, 2611, 3)
class IlastikVersion05(object):
    def __init__(self, h5_handle, image_path=None, prediction=False, skip_image=False):
        self.prediction = prediction
        self.f = h5_handle
        self.image_path = image_path
        self.skip_image = skip_image


    def __getitem__(self, i):
        '''
        Returns `filename, (img, labels, prediction)` for the i'th image
        prediction is None if no prediction was made
        '''
        f = self.f
        dset = f.get('/DataSets/dataItem{:02}'.format(i))

        path = utils.basename(dset.attrs['fileName'].decode('ascii'))
        img = np.array(dset.get('data')) if not self.skip_image else None
        labels = np.array(dset.get('labels/data'))

        if self.prediction:
            prediction = np.array(dset.get('prediction'))
            return path, (img, labels, prediction)
        else:
            return path, (img, labels, None)


    def __iter__(self):
        '''
        Returns `filename, (img, labels, prediction)`
        prediction is None if no prediction was made
        '''
        f = self.f
        for dset_name in f.get('/DataSets').keys():
            i = re.search('[0-9]+$', dset_name).group(0)
            yield self[int(i)]


    def __len__(self):
        return len(self.f.get('DataSets'))



import re
import os

def basename(path):
    '''
    A version of os.path.basename() but it supports windows paths on *nix machines
    '''
    fname =  os.path.basename(path)

    if re.match(r'^[a-zA-Z]:\\', fname):
        # windows path on linux machine (starts with X:\)
        fname = fname.split('\\')[-1]

    return fname




import os
import re
import functools
import warnings
import numpy as np
from bigtiff import Tiff
import pyilastik.utils as utils
from functools import lru_cache

# image shape: (?,?,H,W,C), e.g. (1, 1, 2098, 2611, 3)
# labels shape: (?,?,H,W,1), e.g. (1, 1, 2098, 2611, 1), 0 == unlabeled
# prediction shape: (?,?,H,W,L), e.g. (1, 1, 2098, 2611, 3)


def imread(path):
    '''
    reads tiff image in dimension order zyxc
    '''

    slices = Tiff.memmap_tcz(path)

    img = []
    for C in range(slices.shape[1]):
        img.append(np.stack([s for s in slices[0, C, :]]))
    img = np.stack(img)
    img = np.moveaxis(img, (0, 1, 2, 3), (3, 0, 1, 2))
    return img


def is_overlap(tile_pos, block_pos):
    '''
    checks if plock position overlaps tile position
    '''

    return np.array([block_pos[:, 0] <= tile_pos[:, 1]-1,
                     block_pos[:, 1]-1 >= tile_pos[:, 0]]).all()


def tile_loc_from_slices(slice_list):
    '''
    get optimal tile position and shape for list of slices
    '''
    slice_list = np.array(slice_list)
    assert slice_list.ndim == 3

    lower = np.min(slice_list[:, :, 0], axis=0)
    upper = np.max(slice_list[:, :, 1], axis=0)

    shape = upper - lower
    pos = lower

    return pos, shape


def ndarray2slices(ndarray):

    return [slice(int(start), int(end)) for start, end in ndarray]


def _get_slices_for(p_pos, q_pos, p_shape, q_shape):
    q_lower = p_pos - q_pos
    q_upper = q_lower + p_shape
    q_slice = np.array([q_lower, q_upper]).transpose()
    q_slice[q_slice < 0] = 0
    q_shape_lookup = np.tile(q_shape, (2, 1)).transpose()
    q_slice[q_slice > q_shape_lookup] = q_shape_lookup[q_slice >
                                                       q_shape_lookup]

    return [slice(start, stop) for start, stop in q_slice]


class IlastikStorageVersion01(object):

    def __init__(self, h5_handle, image_path=None, prediction=False,
                 skip_image=False):
        self.prediction = prediction
        self.f = h5_handle
        self.image_path = image_path
        self.skip_image = skip_image

        try:
            version = self.f.get('/Input Data/StorageVersion')[()].decode()
        except AttributeError:
            # for ilastik release > 1.3.0
            version = self.f.get('/Input Data/StorageVersion')[()]
        assert version == '0.2'

    def __iter__(self):
        '''
        Returns `filename, (img, labels, prediction)`

        prediction is None if no prediction was made

        img is a 4D matrix of pixel values in order (Z, Y, X, C) or None

        labels is a 4D matrix in order (Z, Y, X, C) where C size of C dimension
        is always 1 (only one label channel implemented, i.e.
        currently no overlapping label regions supported)
        The shape of labels is equal or smaller to the shape of img
        '''

        for dset_name in self.f.get('/PixelClassification/LabelSets').keys():
            i = re.search('[0-9]+$', dset_name).group(0)

            res = self[int(i)]
            if res is not None:
                yield res

    @functools.lru_cache(1)
    def image_path_list(self):
        '''
        Returns the list of path to the image files on the original file
        system.
        '''
        path_list = []
        for dset_name in self.f.get('/PixelClassification/LabelSets').keys():
            i = re.search('[0-9]+$', dset_name).group(0)
            i = int(i)

            lane = 'lane{:04}'.format(i)
            dset_name = 'labels{:03}'.format(i)

            path = self.f.get(
                '/Input Data/infos/{lane}/Raw Data/filePath'.format(lane=lane))
            path = path[()].decode()

            path_list.append(path)

        return path_list

    def __getitem__(self, i):
        '''
        Returns `filename, (img, labels, prediction)` for the i'th image
        prediction is None if no prediction was made
        '''
        if type(i) == str:
            idx = self.image_path_list().index(i)
            return self[idx]

        f = self.f
        lane = 'lane{:04}'.format(i)
        path = f.get(
            '/Input Data/infos/{lane}/Raw Data/filePath'.format(lane=lane))
        path = path[()].decode()
        original_path = path

        prediction = None  # TODO

        # 1st get the (approximate) labeled image size
        shape = self.shape_of_labelmatrix(i)
        n_dims = len(shape)

        tile_slice = np.array([[0, s] for s in shape])

        labels = self.tile(i, tile_slice)

        msg = 'dimensions of labelmatrix should be 4 (zyxc) or 3 (yxc)'
        assert n_dims in [3, 4], msg

        if n_dims == 3:
            # add z dimension if missing
            labels = np.expand_dims(labels, axis=0)

        if self.skip_image:
            return original_path, (None, labels, prediction)

        fname = utils.basename(path)
        if self.image_path is not None:
            path = os.path.join(self.image_path, fname)

        ilp_path, _ = os.path.split(f.filename)
        if os.path.isfile(path):
            path = path
        elif os.path.isfile(os.path.join(ilp_path, fname)):
            path = os.path.join(ilp_path, fname)
            warnings.warn(
                'Loading file from ilp file path {}'.format(path))
        else:
            warnings.warn(
                '!!! File {} not found. Skipping pixelimage...'.format(path))
            return original_path, (None, labels, prediction)

        img = imread(path)

        # pad labelmatrix to same size as image
        padding = [(0, d2-d1) for d2, d1 in zip(img.shape, labels.shape)]
        padding[-1] = (0, 0)  # set padding for channel axis to 0
        labels = np.pad(labels, padding, mode='constant',
                        constant_values=0)

        return original_path, (img, labels, prediction)

    def n_dims(self, item_index):
        '''
        get nr of dimensions
        3 for yxc
        4 for zyxc
        0 if no labels available
        '''
        slices = self._get_block_slices(item_index)

        if slices.size > 0:
            return slices.shape[1]
        else:
            return 0

    def shape_of_labelmatrix(self, item_index):
        '''
        Label matrix shape is retrieved from label data.

        Order is (Z, Y, X, C) or (Y, X, C) where C size of C dimension
        is always 1 (only one label channel implemented, i.e.
        currently no overlapping label regions supported)

        Label matrix shape does not always equal corresponding image shape.
        Label matrix shape is always smaller or equal to corresponding
        image shape.

        If no labels exist, label matrix shape is 0 in all dimensions.
        '''
        slice_list = self._get_block_slices(item_index)

        if len(slice_list) == 0:
            msg = 'No labels found in Ilastik file - ' +\
                  'cannot approximate image size'
            warnings.warn(msg)

            labelmat_shape = np.zeros((4,), dtype='int')

        else:
            labelmat_shape = np.amax(slice_list[:, :, 1], axis=0)
        return labelmat_shape

    @lru_cache(maxsize=None)
    def _get_blocks(self, item_index):
        dset_name = 'labels{:03}'.format(item_index)
        labelset_str = '/PixelClassification/LabelSets/{}'.format(dset_name)
        return self.f.get(labelset_str).values()

    def _blocks_in_tile(self, item_index, tile_slice):
        '''
        Checks which blocks overlap with the given tile slice.
        Returns a list of booleans (one for each block).
        '''
        block_slices = self._get_block_slices(item_index)

        # We do not check for overlap in the channel dimension (the last
        # dimension). The channel dimension is a dummy dimension with slice
        # values being always [0 1].
        return [is_overlap(tile_slice[:-1, :], block_slice[:-1, :])
                for block_slice in block_slices]

    def tile_for_selected_blocks(self, item_index, block_selection):
        '''
        returns a labelmatrix tile and a corresponding position
        '''
        slices = self._get_block_slices(item_index).copy()
        block_indices = np.nonzero(block_selection)[0]
        pos, shape = tile_loc_from_slices(slices[block_indices])
        labels = np.zeros(shape)

        for i in block_indices:

            s_shifted = slices[i]
            s_shifted[:, 0] = s_shifted[:, 0] - pos
            s_shifted[:, 1] = s_shifted[:, 1] - pos
            s_fmt = ndarray2slices(s_shifted)
            assert (s_shifted >= 0).all()

            labels[s_fmt] = self.load_block_data(item_index, i)

        return pos, labels

    @lru_cache(maxsize=None)
    def load_block_data(self, item_index, block_index):
        blocks = self._get_blocks(item_index)

        for counter, block in enumerate(blocks):
            if counter == block_index:
                return block[()]

    def tile(self, item_index, tile_slice):
        '''
        Order is (Z, Y, X, C) or (Y, X, C) where C size of C dimension
        is always 1 (only one label channel implemented, i.e.
        currently no overlapping label regions supported)
        '''

        sel = self._blocks_in_tile(item_index, tile_slice)

        pos_q = tile_slice[:, 0]
        shape_q = tile_slice[:, 1] - tile_slice[:, 0]
        labels_q = np.zeros(shape_q)

        # return empty label matrix if no blocks in tile
        if not sel:
            return labels_q
        if not np.any(sel):
            return labels_q

        pos_p, labels_p = self.tile_for_selected_blocks(item_index, sel)
        shape_p = labels_p.shape

        q_slice = _get_slices_for(pos_p, pos_q, shape_p, shape_q)
        p_slice = _get_slices_for(pos_q, pos_p, shape_q, shape_p)

        labels_q[q_slice] = labels_p[p_slice]

        return labels_q

    @lru_cache(maxsize=None)
    def _get_block_slices(self, item_index):
        slice_list = []
        for block in self._get_blocks(item_index):
            slices = re.findall('([0-9]+):([0-9]+)',
                                block.attrs['blockSlice'].decode('ascii'))
            slice_list.append(slices)

        return np.array(slice_list).astype('int')

    def __len__(self):
        return len(self.f.get('/PixelClassification/LabelSets').keys())



#!/usr/bin/env python3
'''
Convert Ilastik project files (ilp) to Tiff files.

Usage:
  convert_ilp_to_tif.py <ilp_file> [<image_path>]

'''
from docopt import docopt

import re
import os
import sys

import lib

import numpy as np
import skimage.io

def main(ilp_filepath, image_path):
    output_path = ilp_filepath+'-tiffs'
    try:
        os.mkdir(output_path)
    except:
        pass

    print('Writing output to {}...'.format(output_path))
    for path, (img, labels, _) in lib.read_project(ilp_filepath, image_path=image_path, prediction=False):
        fname = lib.basename(path)

        skimage.io.imsave(os.path.join(output_path, fname+'-img.tif'), img)
        skimage.io.imsave(os.path.join(output_path, fname+'-lbl.tif'), labels)


if __name__ == '__main__':
    args = docopt(__doc__, version='convert_ilp_to_tif 0.1')
    ilp_path = args['<ilp_file>']
    image_path = args['<image_path>']

    try:
        main(ilp_path, image_path)
    except FileNotFoundError as e:
        print('!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!')
        print('!!! Maybe you can fix this error specifying the path to the images using the <image_path> option !!!')
        print('!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!')
        raise e




from unittest import TestCase
import os
import numpy as np
from numpy.testing import assert_array_equal
import pyilastik
from pyilastik.ilastik_storage_version_01 import imread
from bigtiff import Tiff, PlaceHolder
data_path = path = os.path.join(os.path.dirname(__file__),
                                'data/dimensionstest')


val_multi_z = np.array(
    [[[0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.],
      [0., 0., 1., 1., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.],
      [0., 0., 2., 2., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.],
      [0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.],
      [0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.],
      [0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.],
      [0., 0., 0., 0., 0., 0., 0., 0., 1., 1., 0., 0., 0., 0., 0.],
      [0., 0., 0., 0., 0., 0., 0., 0., 2., 2., 0., 0., 0., 0., 0.],
      [0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.],
      [0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.]],
     [[0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.],
      [0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.],
      [0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.],
      [0., 0., 0., 0., 1., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.],
      [0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.],
      [0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.],
      [0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.],
      [0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.],
      [0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.],
      [0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.]]])

val_single_z = val_multi_z[0, :, :]

# swapped label values
val_multi_z_swapped = np.zeros(val_multi_z.shape)
val_multi_z_swapped[val_multi_z == 2] = 1
val_multi_z_swapped[val_multi_z == 1] = 2
val_single_z_swapped = val_multi_z_swapped[0, :, :]

class TestImgImport(TestCase):

    def test_imread(self):

        img_path = os.path.join(path, 'x15_y10_z1_c1.tif')
        img = imread(img_path)
        self.assertEqual(img.shape, (1, 10, 15, 1))

        img_path = os.path.join(path, 'x15_y10_z1_c2.tif')
        img = imread(img_path)
        self.assertEqual(img.shape, (1, 10, 15, 2))

        img_path = os.path.join(path, 'x15_y10_z2_c4.tif')
        img = imread(img_path)
        self.assertEqual(img.shape, (2, 10, 15, 4))

        img_path = os.path.join(path, 'x15_y10_z2_c1.tif')
        img = imread(img_path)
        self.assertEqual(img.shape, (2, 10, 15, 1))

        img_path = os.path.join(path, 'x15_y10_z1_rgb.tif')
        img = imread(img_path)
        self.assertEqual(img.shape, (1, 10, 15, 3))

        img_path = os.path.join(path, 'x15_y10_z2_rgb.tif')
        img = imread(img_path)
        self.assertEqual(img.shape, (2, 10, 15, 3))

    def test_single_channel_single_z(self):

        p = os.path.join(path, 'x15_y10_z1_c1_classes2.ilp')
        ilp = pyilastik.read_project(p, skip_image=False)
        (_, (img, labels, _)) = ilp['x15_y10_z1_c1.tif']

        self.assertEqual(img.shape, (1, 10, 15, 1))
        self.assertEqual(labels.shape, (1, 10, 15, 1))
        assert_array_equal(labels[0, :, :, 0], val_single_z)

    def test_single_channel_multi_z(self):

        p = os.path.join(path, 'x15_y10_z2_c1_classes2.ilp')
        ilp = pyilastik.read_project(p, skip_image=False)
        (_, (img, labels, _)) = ilp['x15_y10_z2_c1.tif']

        self.assertEqual(img.shape, (2, 10, 15, 1))
        self.assertEqual(labels.shape, (2, 10, 15, 1))
        assert_array_equal(labels[:, :, :, 0], val_multi_z)

    def test_multi_channel_multi_z(self):

        p = os.path.join(path, 'x15_y10_z2_c4_classes2.ilp')
        ilp = pyilastik.read_project(p, skip_image=False)
        (_, (img, labels, _)) = ilp['x15_y10_z2_c4.tif']

        self.assertEqual(img.shape, (2, 10, 15, 4))
        self.assertEqual(labels.shape, (2, 10, 15, 1))
        assert_array_equal(labels[:, :, :, 0], val_multi_z)

    def test_rgb_single_z(self):

        p = os.path.join(path, 'x15_y10_z1_rgb_classes2.ilp')
        ilp = pyilastik.read_project(p, skip_image=False)
        (_, (img, labels, _)) = ilp['x15_y10_z1_rgb.tif']

        self.assertEqual(img.shape, (1, 10, 15, 3))
        self.assertEqual(labels.shape, (1, 10, 15, 1))
        assert_array_equal(labels[0, :, :, 0], val_single_z_swapped)

    def test_rgb_multi_z(self):

        p = os.path.join(path, 'x15_y10_z2_rgb_classes2.ilp')
        ilp = pyilastik.read_project(p, skip_image=False)
        (_, (img, labels, _)) = ilp['x15_y10_z2_rgb.tif']

        self.assertEqual(img.shape, (2, 10, 15, 3))
        self.assertEqual(labels.shape, (2, 10, 15, 1))
        assert_array_equal(labels[:, :, :, 0], val_multi_z)

    def test_large_dat(self):

        localpath = os.path.join(path, '../purkinjetest')
        p = os.path.join(localpath, 'ilastik-1.2.2post1mac.ilp')
        ilp = pyilastik.read_project(p, skip_image=False,
                                     image_path=os.path.join(localpath,
                                                             'images'))
        (_, (img, labels, _)) = ilp['images/769_cerebellum_5M41_subset_1.tif']

        self.assertEqual(img.shape, (1, 684, 1047, 3))
        self.assertEqual(labels.shape, (1, 684, 1047, 1))

        val = np.array([[2., 2., 2., 2.],
                        [0., 2., 2., 2.],
                        [0., 0., 2., 2.],
                        [0., 0., 4., 4.],
                        [4., 4., 4., 4.]])

        assert_array_equal(val, labels[0, 212:217, 309:313, 0])



import h5py
from pyilastik.ilastik_version_05 import IlastikVersion05
# for ilastik 1.2
from pyilastik.ilastik_storage_version_01 import IlastikStorageVersion01


class UnknownIlastikVersion(Exception):
    pass

# image shape: (?,?,H,W,C), e.g. (1, 1, 2098, 2611, 3)
# labels shape: (?,?,H,W,1), e.g. (1, 1, 2098, 2611, 1), 0 == unlabeled
# prediction shape: (?,?,H,W,L), e.g. (1, 1, 2098, 2611, 3)


def read_project(ilastik_file, image_path=None, prediction=False,
                 skip_image=False):
    f = h5py.File(ilastik_file, 'r')
    version = f.get('/PixelClassification/StorageVersion')
    if version is not None:
        try:
            version = version[()].decode()
        except AttributeError:
            # for ilastik releases >1.3.0
            version = version[()]

        # for ilastik release >1.2 (storage version 0.1)
        if version == '0.1':
            return IlastikStorageVersion01(f, image_path, prediction,
                                           skip_image)

    if version is None:
        # old ilastik version 0.5
        version = f.get('ilastikVersion')[()]
        if version == 0.5:
            return IlastikVersion05(f, image_path, prediction, skip_image)
    msg = ('Unknown ilastik version v{}. I only understand ilastik v0.5 and' +
           'ilastik storage version v0.1 (ilastik v1.2.0)!').format(format)
    raise UnknownIlastikVersion(msg)



import os

from setuptools import setup, find_packages

try:  # for pip >= 10
    from pip._internal.req import parse_requirements
except ImportError:  # for pip <= 9.0.3
    from pip.req import parse_requirements


requirements_txt = os.path.join(os.path.dirname(__file__), 'requirements.txt')
install_reqs = parse_requirements(requirements_txt, session=False)
reqs = [str(ir.req) for ir in install_reqs]

def readme():
    README_md = os.path.join(os.path.dirname(__file__), 'README.md')
    with open(README_md) as f:
        return f.read()


setup(name='pyilastik',
      version='0.0.4',
      description='Read ilastik labels in python',
      author='Manuel Schoelling',
      author_email='manuel.schoelling@dzne.de',
      include_package_data=True,
      packages=find_packages(exclude=['contrib', 'docs', 'tests*']),
      zip_safe=False,
      install_requires=reqs,
      test_suite='nose.collector',
      tests_require=['nose', 'coverage', 'nose-timer', 'nose-deadline'])
